class CarolApiResponseException(Exception):
    """Custom Exception to handle Exception on Carol API Calls

    Args:
        Exception (Exception): SuperClass
    """
    pass

class InvalidToken(Exception):

    """Custom Exception to handle Invalid token
    Args:
        Exception (Exception): SuperClass
    """
    pass

